use crate::ast::named::Named;
use crate::ast::CompilationUnit;
use crate::diagnostic::DmDiagnostic;
use crate::name_analysis::gather::gather_compilation_unit;
use crate::name_analysis::resolve::resolve_compilation_unit;
use crate::name_analysis::symbol_table::SymbolTable;

mod fqn_context;
mod gather;
mod resolve;
pub mod symbol;
pub mod symbol_table;

pub fn analyze_names(
    compilation_units: &mut Vec<CompilationUnit>,
    symbol_table: &mut SymbolTable,
) -> Vec<DmDiagnostic> {
    let mut diagnostics = vec![];

    // gather symbols
    for compilation_unit in compilation_units.iter_mut() {
        gather_compilation_unit(compilation_unit, symbol_table, &mut diagnostics);
    }

    // resolve symbols
    for compilation_unit in compilation_units.iter_mut() {
        resolve_compilation_unit(compilation_unit, symbol_table, &mut diagnostics);
    }

    diagnostics.into()
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::ast::build::build_ast;
    use crate::parser::{DeimosParser, Rule};
    use crate::std_core::add_std_core_symbols;
    use codespan_reporting::files::SimpleFiles;
    use codespan_reporting::term;
    use codespan_reporting::term::termcolor::{ColorChoice, StandardStream};
    use indoc::indoc;
    use pest::Parser;
    use std::collections::HashMap;

    fn assert_no_diagnostics(sources: HashMap<&str, &str>, symbol_table: &mut SymbolTable) {
        let mut files = SimpleFiles::new();
        let mut compilation_units = vec![];

        for (file_name, source) in sources {
            let file_id = files.add(file_name, source);
            let parse_result = DeimosParser::parse(Rule::CompilationUnit, source);
            if let Err(err) = &parse_result {
                panic!("{}", err);
            }
            let mut pairs = parse_result.unwrap();
            if pairs.as_str().trim() != source.trim() {
                panic!("Parsing did not consume entire input.");
            }

            compilation_units.push(build_ast(file_name, file_id, pairs.next().unwrap()))
        }

        let diagnostics = analyze_names(&mut compilation_units, symbol_table);

        if !diagnostics.is_empty() {
            let writer = StandardStream::stderr(ColorChoice::Always);
            let config = term::Config::default();

            for diagnostic in &diagnostics {
                term::emit(&mut writer.lock(), &config, &files, &diagnostic).unwrap();
            }

            eprintln!("{}", symbol_table);
            panic!("Diagnostics was not empty!");
        }

        for compilation_unit in &compilation_units {
            dbg!(compilation_unit);
        }
    }

    #[test]
    fn params_seen() {
        let sources: HashMap<&str, &str> = HashMap::from([(
            "main.dm",
            indoc! {"
            fn main(args: Array<String>) {
                let x = args;
            }"},
        )]);

        assert_no_diagnostics(sources, &mut SymbolTable::new());
    }

    #[test]
    fn two_files() {
        let sources: HashMap<&str, &str> = HashMap::from([
            (
                "main.dm",
                indoc! {"
                use test::Greeter;
            "},
            ),
            (
                "deps.dm",
                indoc! {"
                ns test;

                pub class Greeter {}
            "},
            ),
        ]);

        assert_no_diagnostics(sources, &mut SymbolTable::new());
    }

    #[test]
    fn sees_std_core_println() {
        let sources: HashMap<&str, &str> = HashMap::from([(
            "main.dm",
            indoc! {"
                fn main(args: Array<String>) {
                    println(args)
                }
            "},
        )]);

        let mut symbol_table = SymbolTable::new();
        add_std_core_symbols(&mut symbol_table).expect("Failed to add std::core symbols.");
        assert_no_diagnostics(sources, &mut symbol_table);
    }
}
