use crate::ast::named::Named;
use crate::ast::*;
use crate::diagnostic::DmDiagnostic;
use crate::name_analysis::symbol_table::SymbolTable;
use codespan_reporting::diagnostic::{Diagnostic, Label};
use std::cell::RefCell;
use std::range::Range;
use std::rc::Rc;

fn resolve_fully_qualified_name(
    fully_qualified_name: &mut FullyQualifiedName,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    let lookup_result = symbol_table.lookup(
        fully_qualified_name.name().as_ref(),
        fully_qualified_name.scope_id().expect(&format!(
            "FullyQualifiedName has no scope_id set: {:?}",
            fully_qualified_name
        )),
    );
    match lookup_result {
        Ok(symbol) => {
            fully_qualified_name.set_symbol(symbol.clone());
        }
        Err(e) => diagnostics.push(
            Diagnostic::error()
                .with_message(format!(
                    "No symbol with name '{}' found in current scope.",
                    fully_qualified_name.name()
                ))
                .with_label(Label::primary(
                    fully_qualified_name.file_id,
                    fully_qualified_name.range,
                )),
        ),
    }
}

fn resolve_type_use(
    type_use: &mut TypeUse,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    match type_use {
        TypeUse::Primitive(primitive_type_use) => {
            match primitive_type_use {
                PrimitiveTypeUse::Array(generic_arguments_opt) => {
                    if let Some(generic_arguments) = generic_arguments_opt {
                        resolve_generic_arguments(
                            generic_arguments,
                            symbol_table,
                            diagnostics,
                        );
                    }
                }
                _ => {}
            }
        }
        TypeUse::InterfaceOrClass(interface_or_class_type_use) => {
            resolve_interface_or_class_type_use(
                interface_or_class_type_use,
                symbol_table,
                diagnostics,
            );
        }
        TypeUse::Tuple(tuple_type_use) => {
            resolve_tuple_type_use(tuple_type_use, symbol_table, diagnostics);
        }
        TypeUse::Function(function_type_use) => {
            resolve_function_type_use(function_type_use, symbol_table, diagnostics);
        }
    }
}

fn resolve_interface_or_class_type_use(
    interface_or_class_type_use: &mut InterfaceOrClassTypeUse,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_fully_qualified_name(
        &mut interface_or_class_type_use.fqn,
        symbol_table,
        diagnostics,
    );
    resolve_generic_arguments(
        &mut interface_or_class_type_use.generics,
        symbol_table,
        diagnostics,
    );
}

fn resolve_tuple_type_use(
    tuple_type_use: &mut TupleTypeUse,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_tuple_arguments(&mut tuple_type_use.arguments, symbol_table, diagnostics);
}

fn resolve_function_type_use(
    function_type_use: &mut FunctionTypeUse,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_parameters(&mut function_type_use.parameters, symbol_table, diagnostics);
    resolve_return_type(
        &mut function_type_use.return_type,
        symbol_table,
        diagnostics,
    );
}

fn resolve_generic_arguments(
    generic_arguments: &mut GenericArguments,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for generic_argument in &mut generic_arguments.0 {
        resolve_type_use(generic_argument, symbol_table, diagnostics);
    }
}

fn resolve_tuple_arguments(
    tuple_type_use: &mut TupleArguments,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for type_use in &mut tuple_type_use.0 {
        resolve_type_use(type_use, symbol_table, diagnostics);
    }
}

fn resolve_implements_list(
    implements_list: &mut ImplementsList,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    todo!()
}

fn resolve_parameters(
    parameters: &mut Parameters,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for parameter in &mut parameters.0 {
        resolve_type_use(&mut parameter.type_use, symbol_table, diagnostics);
    }
}

fn resolve_return_type(
    return_type: &mut ReturnType,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_type_use(&mut return_type.declared_type, symbol_table, diagnostics);
    resolve_references(&mut return_type.references, symbol_table, diagnostics);
}

fn resolve_references(
    references: &mut References,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for reference in &mut references.0 {
        todo!()
    }
}

fn handle_use_statement_identifier(
    identifier: &Rc<RefCell<Identifier>>,
    base_name: &str,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
    file_id: usize,
    error_range: Range<usize>,
) {
    let borrowed_identifier = identifier.borrow();
    let declared_name = borrowed_identifier.name().to_string();
    let fqn = format!("{}::{}", base_name, &declared_name);
    let lookup_result =
        symbol_table.lookup_usable_by_fqn(&fqn, borrowed_identifier.scope_id().unwrap());

    match lookup_result {
        Ok(referenced_symbol) => match borrowed_identifier.saved_symbol() {
            Some(saved_symbol) => match saved_symbol {
                SavedSymbol::Linking(symbol) => {
                    let use_statement_symbol = symbol.unwrap_use_statement_symbol();
                    use_statement_symbol
                        .borrow_mut()
                        .set_referenced_symbol(referenced_symbol);
                }
                _ => panic!("Expected to find SavedSymbol::Linking."),
            },
            None => panic!("Expected to find saved_symbol"),
        },
        Err(_) => {
            diagnostics.push(
                Diagnostic::error()
                    .with_message(&format!("Unable to find symbol '{}'.", fqn))
                    .with_label(Label::primary(file_id, error_range)),
            );
        }
    }
}

fn resolve_use_statement(
    use_statement: &mut UseStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    let base_name = use_statement.base_name().to_string();

    match &use_statement.last {
        UseStatementLast::Identifier(identifier) => {
            handle_use_statement_identifier(
                identifier,
                &base_name,
                symbol_table,
                diagnostics,
                use_statement.file_id,
                use_statement.range,
            );
        }
        UseStatementLast::Identifiers(identifiers) => {
            for identifier in identifiers {
                handle_use_statement_identifier(
                    identifier,
                    &base_name,
                    symbol_table,
                    diagnostics,
                    use_statement.file_id,
                    identifier.borrow().range,
                )
            }
        }
        UseStatementLast::Star => todo!(),
    }
}

pub(super) fn resolve_compilation_unit(
    compilation_unit: &mut CompilationUnit,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for use_statement in &mut compilation_unit.use_statements {
        resolve_use_statement(use_statement, symbol_table, diagnostics);
    }
    for declaration in &mut compilation_unit.declarations {
        resolve_module_level_declaration(declaration, symbol_table, diagnostics);
    }
}

pub(super) fn resolve_module_level_declaration(
    declaration: &mut ModuleLevelDeclaration,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    use crate::ast::ModuleLevelDeclaration::*;
    match declaration {
        Function(function_definition) => {
            resolve_function_definition(function_definition, symbol_table, diagnostics)
        }
        PlatformFunction(platform_function_declaration) => resolve_platform_function_declaration(
            platform_function_declaration,
            symbol_table,
            diagnostics,
        ),
        Class(class_declaration) => {
            // todo
        }
        _ => todo!(),
    }
}

fn resolve_function_definition(
    function_definition: &mut FunctionDefinition,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_parameters(
        &mut function_definition.parameters,
        symbol_table,
        diagnostics,
    );
    resolve_return_type(
        &mut function_definition.return_type,
        symbol_table,
        diagnostics,
    );
    resolve_function_body(&mut function_definition.body, symbol_table, diagnostics);
}

fn resolve_platform_function_declaration(
    platform_function_declaration: &mut PlatformFunctionDeclaration,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_parameters(
        &mut platform_function_declaration.parameters,
        symbol_table,
        diagnostics,
    );
    resolve_return_type(
        &mut platform_function_declaration.return_type,
        symbol_table,
        diagnostics,
    );
}

fn resolve_function_body(
    function_body: &mut FunctionBody,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    use crate::ast::FunctionBody::*;
    match function_body {
        Block(block) => resolve_block(block, symbol_table, diagnostics),
        _ => todo!(),
    }
}

fn resolve_block(
    block_statement: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for statement in block_statement.statements.iter_mut() {
        resolve_statement(statement, symbol_table, diagnostics);
    }
    if let Some(expression) = block_statement.expression.as_mut() {
        resolve_expression(expression, symbol_table, diagnostics);
    }
}

fn resolve_statement(
    statement: &mut Statement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    use crate::ast::Statement::*;
    match statement {
        BlockStatement(block) => resolve_block(block, symbol_table, diagnostics),
        VariableDeclarationStatement(variable_declaration) => {
            resolve_variable_declaration(variable_declaration, symbol_table, diagnostics)
        }
        AssignStatement(assign_statement) => {
            resolve_assign_statement(assign_statement, symbol_table, diagnostics)
        }
        CallStatement(call_statement) => {
            resolve_call_statement(call_statement, symbol_table, diagnostics)
        }
        _ => todo!(),
    }
}

fn resolve_variable_declaration(
    variable_declaration: &mut VariableDeclarationStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    if let Some(initializer) = &mut variable_declaration.initializer {
        resolve_expression(initializer, symbol_table, diagnostics)
    }
}

fn resolve_assign_statement(
    assign_statement: &mut AssignStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut assign_statement.lhs, symbol_table, diagnostics);
    resolve_expression(&mut assign_statement.rhs, symbol_table, diagnostics);
}

fn resolve_call_statement(
    call_statement: &mut CallStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut call_statement.0, symbol_table, diagnostics)
}

fn resolve_expression(
    expression: &mut Expression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    use crate::ast::Expression::*;
    match expression {
        Ternary(ternary_expression) => {
            resolve_ternary_expression(ternary_expression, symbol_table, diagnostics);
        }
        Binary(binary_expression) => {
            resolve_binary_expression(binary_expression, symbol_table, diagnostics);
        }
        UnaryPrefix(prefix_expression) => {
            resolve_prefix_expression(prefix_expression, symbol_table, diagnostics);
        }
        UnarySuffix(suffix_expression) => {
            resolve_suffix_expression(suffix_expression, symbol_table, diagnostics);
        }
        Call(call_expression) => {
            resolve_call_expression(call_expression, symbol_table, diagnostics)
        }
        ObjectAccess(object_access) => {
            resolve_object_access(object_access, symbol_table, diagnostics);
        }
        FullyQualifiedName(fqn) => resolve_fully_qualified_name(fqn, symbol_table, diagnostics),
        Literal(literal) => {
            resolve_literal(literal, symbol_table, diagnostics);
        }
        Closure(closure) => {
            resolve_closure(closure, symbol_table, diagnostics);
        }
    }
}

fn resolve_ternary_expression(
    ternary_expression: &mut TernaryExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut ternary_expression.condition, symbol_table, diagnostics);
    resolve_expression(
        &mut ternary_expression.true_expression,
        symbol_table,
        diagnostics,
    );
    resolve_expression(
        &mut ternary_expression.false_expression,
        symbol_table,
        diagnostics,
    );
}

fn resolve_binary_expression(
    binary_expression: &mut BinaryExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut binary_expression.left, symbol_table, diagnostics);
    resolve_expression(&mut binary_expression.right, symbol_table, diagnostics);
}

fn resolve_prefix_expression(
    prefix_expression: &mut PrefixExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut prefix_expression.expression, symbol_table, diagnostics);
}

fn resolve_suffix_expression(
    suffix_expression: &mut SuffixExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut suffix_expression.expression, symbol_table, diagnostics);
}

fn resolve_call_expression(
    call_expression: &mut CallExpression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut call_expression.callee, symbol_table, diagnostics);
    if let Some(turbo_fish) = &mut call_expression.turbo_fish {
        resolve_turbo_fish(turbo_fish, symbol_table, diagnostics);
    }
    resolve_call_arguments(&mut call_expression.arguments, symbol_table, diagnostics);
}

fn resolve_turbo_fish(
    turbo_fish: &mut TurboFish,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    todo!()
}

fn resolve_call_arguments(
    call_arguments: &mut CallArguments,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    for argument in &mut call_arguments.0 {
        resolve_call_argument(argument, symbol_table, diagnostics);
    }
}

fn resolve_call_argument(
    call_argument: &mut CallArgument,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    resolve_expression(&mut call_argument.0, symbol_table, diagnostics);
}

fn resolve_closure(
    closure: &mut Closure,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    todo!()
}

fn resolve_object_access(
    object_access: &mut ObjectAccess,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    todo!()
}

fn resolve_literal(
    literal: &mut Literal,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    match literal {
        Literal::DString(d_string) => resolve_d_string(d_string, symbol_table, diagnostics),
        Literal::BacktickString(d_string) => resolve_d_string(d_string, symbol_table, diagnostics),
        _ => {}
    }
}

fn resolve_d_string(
    d_string: &mut DString,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut Vec<DmDiagnostic>,
) {
    todo!()
}
